//------------------------------------------------------------------------------
/*
    This file is part of Bottle2D.

    Bottle2D is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Bottle2D is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Bottle2D.  If not, see <http://www.gnu.org/licenses/>.
*/
//------------------------------------------------------------------------------
#include "Object.hpp"
//------------------------------------------------------------------------------
#include "Group.hpp"
#include "Screen.hpp"
//------------------------------------------------------------------------------
#include "MyException.hpp"
//------------------------------------------------------------------------------
#include <cmath>
#include <string>
#include <iostream>
//------------------------------------------------------------------------------
#include "AllObj.hpp"
//------------------------------------------------------------------------------
namespace En
{
//------------------------------------------------------------------------------
using namespace std;
//------------------------------------------------------------------------------
Object::Object(Group *parent, const std::string &name) : m_parent(parent), m_name(name), m_speed(Vector2D(0, 0)), m_angularVelocity(0.f), m_pos(Vector2D(0, 0)), m_rotation(0.f)
{
    // On remplie les m_global*
    RefreshGlobalRotatedPos();
    #if DEBUG == 1
    GetPtrInfo()->objName = GetName();
    #endif
}
//------------------------------------------------------------------------------
Object::~Object()
{
    /*if(HaveReferencedPtr())
    {
        cerr << "Warning: object " << GetName() << " is still referenced while deleted." << endl;
    }*/
}
//------------------------------------------------------------------------------
Group *Object::GetParentRaw()
{
    return m_parent;
}
//------------------------------------------------------------------------------
void Object::OnChangeScreen(Screen *backScreen, Screen *newScreen)
{

}
//------------------------------------------------------------------------------
void Object::SetParent(Group *parent)
{
    m_parent = parent;
}
//------------------------------------------------------------------------------
/** Create a clone of the object and return an handler to the clone. */
luabind::object Object::Clone() const
{
    return CloneToGroup(NULL);
}
//------------------------------------------------------------------------------
/** Create a clone of the object, change the parent and return an handler to the clone. */
luabind::object Object::CloneToGroup(Group *newParent) const
{
    return CastObject(GetRawCloneToGroup(newParent));
}

//------------------------------------------------------------------------------
/** Set user data */
void Object::SetUserData(const std::string &id, const luabind::object &data)
{
    m_userData[id] = data;
}
//------------------------------------------------------------------------------
/** Get user data */
luabind::object Object::GetUserData(const std::string &id)
{
    std::map<std::string, luabind::object>::iterator it = m_userData.find(id);

    if(it != m_userData.end())
    {
        return it->second;
    }
    else
    {
        MY_THROW("Object::GetUserData : Key '" << id << "' not found in object '" << GetName() << "' !");
    }
}
//------------------------------------------------------------------------------
/** Return true if the user data exists, false otherwise */
bool Object::UserDataExists(const std::string &id)
{
    return(m_userData.find(id) != m_userData.end());
}
//------------------------------------------------------------------------------
/** Get the object's parent (can be a Group or a Screen) */
luabind::object Object::GetParent()
{
    return CastObject(GetParentRaw());
}
//------------------------------------------------------------------------------
/** Get the screen (main parent) of this object */
ProtectedPtr<Screen> Object::GetParentScreen()
{
    return GetParentScreenRaw();
}
//------------------------------------------------------------------------------
Screen *Object::GetParentScreenRaw()
{
    if(GetParentRaw() == NULL)
    {
        if(GetType() == OBJ_TYPE_DRAWABLE_SCREEN)
        {
            return static_cast<Screen*>(this);
        }
        else
        {
            MY_THROW("Main parent of '" << GetName() << "' isn't a Screen");
        }
    }
    if(GetParentRaw()->GetType() == OBJ_TYPE_DRAWABLE_SCREEN)
    {
        return static_cast<Screen*>(GetParentRaw());
    }
    else
    {
        return(GetParentRaw()->GetParentScreenRaw());
    }
}

//------------------------------------------------------------------------------
std::string Object::GetName() const
{
    return m_name;
}

//------------------------------------------------------------------------------
void Object::GetPolarPosition(float &dist, float &angle) const
{
    dist = GetPosition().GetSize();
    if(dist == 0)
    {
        angle = 0;
        return;
    }
    angle = - 2 * atan(GetPosition().y / (GetPosition().x + dist)); // Cf http://fr.wikipedia.org/wiki/Coordonn%C3%A9es_polaires

    return;
}
//------------------------------------------------------------------------------
void Object::SetPolarPosition(float dist, float angle)
{
    if(dist != 0)
    {
        SetPosition(Vector2D(dist * cos(angle), - dist * sin(angle)));
    }
    else
    {
        SetPosition(Vector2D(0, 0));
    }
}
//------------------------------------------------------------------------------
void Object::Destroy()
{
    if(m_parent != NULL)
    {
        ChangeParent(NULL);
        delete this;
    }
    else
    {
        MY_THROW("Object::Destroy(): No parent !!")
    }
}
//------------------------------------------------------------------------------
void Object::SetName(const std::string &str)
{
    m_name = str;
    #if DEBUG == 1
    GetPtrInfo()->objName = GetName();
    #endif
}
//------------------------------------------------------------------------------
Vector2D Object::GetGlobalPosition() const
{
    /*if(m_parent != NULL)
    {
        return m_parent->GetGlobalPosition() + m_pos;
    }
    else
    {
        return m_pos;
    }*/

    return m_globalRotatedPos;
}
//------------------------------------------------------------------------------
/** Refresh the global rotated pos and propagate it to children objects */
void Object::PropagateGlobalRotatedPos(const Vector2D &parentPos, float parentAngle)
{
    // On applique sur cet objet
    if(parentAngle == 0) // Dans 80% des cas, pas besoin de faire de calculs
    {
        m_globalRotatedPos = parentPos + GetPosition();
    }
    else
    {
        m_globalRotatedPos = parentPos + PointRotate(Vector2D(0, 0), GetPosition(), parentAngle);
    }

    m_globalRotation = parentAngle + GetRotation(); // La rotation d'un objet ne modifie pas sa position => Elle modifie l'objet affiché ou les childs des groupes

}
//------------------------------------------------------------------------------

/** Refresh the global rotated pos and propagate it to children objects, starting in this object */
void Object::RefreshGlobalRotatedPos()
{
    if(m_parent != NULL)
    {
        PropagateGlobalRotatedPos(m_parent->GetGlobalPosition(), m_parent->GetGlobalRotation());
    }
    else
    {
        PropagateGlobalRotatedPos(Vector2D(0, 0), 0);
    }
}
//------------------------------------------------------------------------------
void Object::SetGlobalPosition(const Vector2D &pos)
{
    if(m_parent != NULL)
    {
        SetPosition(pos - m_parent->GetGlobalPosition());
    }
    else
    {
        SetPosition(pos);
    }
}
//------------------------------------------------------------------------------
float Object::GetGlobalRotation() const
{
    return m_globalRotation;
    /*if(m_parent != NULL)
    {
        return m_parent->GetGlobalRotation() + m_rotation;
    }
    else
    {
        return m_rotation;
    }*/
}
//------------------------------------------------------------------------------
void Object::SetGlobalRotation(float rotation)
{
    if(m_parent != NULL)
    {
        SetRotation(rotation - m_parent->GetGlobalRotation());
    }
    else
    {
        SetRotation(rotation);
    }
}
//------------------------------------------------------------------------------
Vector2D Object::GetPosition() const
{
    return m_pos;
}
//------------------------------------------------------------------------------
void Object::SetPosition(const Vector2D &pos)
{
    m_pos = pos;

    // Refresh => Super important !
    RefreshGlobalRotatedPos();
}
//------------------------------------------------------------------------------
float Object::GetRotation() const
{
    return m_rotation;
}
//------------------------------------------------------------------------------
void Object::SetRotation(float rotation)
{
    m_rotation = floatingModulo(rotation, float(2.f * M_PI));

    // Refresh => Super important !
    RefreshGlobalRotatedPos();
}
//------------------------------------------------------------------------------
void Object::Rotate(float angle)
{
    if(angle != 0)
    {
        SetRotation(GetRotation() + angle);
    }
}
//------------------------------------------------------------------------------
void Object::SetSpeed(const Vector2D &speed)
{
    m_speed = speed;
}
//------------------------------------------------------------------------------
Vector2D Object::GetSpeed() const
{
    return m_speed;
}
//------------------------------------------------------------------------------
void Object::SetAngularVelocity(float speed)
{
    m_angularVelocity = speed;
}
//------------------------------------------------------------------------------
float Object::GetAngularVelocity() const
{
    return m_angularVelocity;
}
//------------------------------------------------------------------------------
void Object::Move(const Vector2D &mvt)
{
    if(mvt != Vector2D(0, 0))
    {
        SetPosition(GetPosition() + mvt);
    }
}
//------------------------------------------------------------------------------
void Object::ProceedMvt(float time)
{
    Move(GetSpeed() * time);
    Rotate(GetAngularVelocity() * time);
}
//------------------------------------------------------------------------------
long Object::GetUniqueId() const
{
    return reinterpret_cast<long>(this);
}

//------------------------------------------------------------------------------
luabind::object Object::CastObject(Object *obj) const
{
    if(obj == NULL)
    {
        MY_THROW("Object::CastObject() : Obj is NULL in group '" << GetName() << "'!");
    }

    luabind::object luaObj;
    switch(obj->GetType())
    {
        #define X(typeName, typeId) \
                case typeId: \
                    luaObj = luabind::object(Global::GetLuaState(), ProtectedPtr<typeName>(static_cast<typeName*>(obj))); \
                    break;
            X_OBJ_LIST
        #undef X
        default:
            MY_THROW("Object::CastObject() : Unknown type << '" << obj->GetType() << "' in group '" << GetName() << "'!");
            break;
    }

    return luaObj;
}
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
} // namespace En
